"""
downloader handles downloading, writing to disc and providing feedback through
a progress bar
"""

import aiofiles
import aiohttp
import os
import tqdm
import typing
import urllib.parse
from .utils import *
from .errors import *


async def download(
        url: str,
        headers: dict[str, str],
        session: aiohttp.ClientSession,
        chunk_size: int,
) -> typing.Union[dict[str, typing.Union[str, int]], bytes]:
    """
    performs http get request for given url and returns metadata and chunked
    data
    """

    async with session.get(url=url, headers=headers) as response:
        metadata = {
            "content_type": response.content_type,
            "content_length": response.content_length,
            "status": response.status,
        }

        yield metadata

        if not response.ok:
            return

        async for data in response.content.iter_chunked(chunk_size):
            yield data


def progress_bar(file_name: str, total: typing.Optional[int]) -> tqdm.std.tqdm:
    """returns tqdm progress bar ready for use"""

    desc = f"{info_banner()} Downloading: {format_user_submitted(file_name)}"

    if total is not None:
        bar_format = "{desc} |\033[1;33m{bar}\033[0m| " + \
            "\033[1m{percentage:06.2f}%\033[0m " + \
            "\033[1;32m{elapsed}\033[0m " + \
            "\033[7;37m{rate_fmt}\033[0m"

        return tqdm.tqdm(
            unit="B",
            unit_scale=True,
            desc=desc,
            total=total,
            bar_format=bar_format
        )

    bar_format = "{desc} \033[1;32m{elapsed}\033[0m \033[7;37m{rate_fmt}\033[0m"

    return tqdm.tqdm(
        unit="B",
        unit_scale=True,
        desc=desc,
        bar_format=bar_format
    )


async def worker(idx: int, url: str, session: aiohttp.ClientSession):
    """handles downloading, writing to disc and updating progress bar"""

    _CHUNK_SIZE = 60*1024

    parsed_url = urllib.parse.urlparse(url)
    file_name = parsed_url.path.split("/")[-1]
    file_name = urllib.parse.unquote(file_name)

    if file_name == "":
        file_name = f"file_{idx+1}"

    headers = {}
    resume_download = True if os.path.exists(file_name) else False
    if resume_download:
        file_size = os.stat(file_name).st_size
        headers["Range"] = f"bytes={file_size}-"

    download_generator = download(
        url=url,
        headers=headers,
        session=session,
        chunk_size=_CHUNK_SIZE
    )

    metadata = await download_generator.asend(None)

    if metadata["status"] == 416:
        bar = progress_bar(file_name=file_name, total=file_size)
        bar.update(file_size)
        return

    elif metadata["status"] >= 400:
        raise KetterHTTPError(
            f"Response status {metadata['status']}, {HTTP_CODES[metadata['status']]}")

    really_resume_download = resume_download and metadata["status"] == 206

    total = metadata["content_length"]
    if really_resume_download:
        total += file_size

    bar = progress_bar(file_name=file_name, total=total)

    if really_resume_download:
        bar.update(file_size)

    open_mode = "ab" if really_resume_download else "wb"
    async with aiofiles.open(file_name, open_mode) as file:
        async for data in download_generator:
            await file.write(data)
            bar.update(len(data))



"""utils contains convenient helper functions and variables"""

from .colour import *

HTTP_CODES = {
    100: 'Continue', 101: 'Switching Protocols', 102: 'Processing',
    103: 'Early Hints', 200: 'OK', 201: 'Created', 202: 'Accepted',
    203: 'Non-Authoritative Information', 204: 'No Content', 205:
    'Reset Content', 206: 'Partial Content', 207: 'Multi-Status', 208:
    'Already Reported', 226: 'IM Used', 300: 'Multiple Choices', 301:
    'Moved Permanently', 302: 'Found', 303: 'See Other', 304: 'Not Modified',
    305: 'Use Proxy', 306: 'Switch Proxy', 307: 'Temporary Redirect',
    308: 'Permanent Redirect', 400: 'Bad Request', 401:
    'Unauthorized', 402: 'Payment Required', 403: 'Forbidden', 404:
    'Not Found', 405: 'Method Not Allowed', 406: 'Not Acceptable', 407:
    'Proxy Authentication Required', 408: 'Request Timeout', 409:
    'Conflict', 410: 'Gone', 411: 'Length Required', 412: 'Precondition Failed',
    413: 'Payload Too Large', 414: 'URI Too Long', 415:
    'Unsupported Media Type', 416: 'Range Not Satisfiable', 417:
    'Expectation Failed', 418: "I'm a teapot", 421: 'Misdirected Request',
    422: 'Unprocessable Entity', 423: 'Locked',
    424: 'Failed Dependency', 425: 'Too Early', 426: 'Upgrade Required', 428:
    'Precondition Required', 429: 'Too Many Requests',
    431: 'Request Header Fields Too Large', 451: 'Unavailable For Legal Reasons',
    500: 'Internal Server Error', 501: 'Not Implemented', 502: 'Bad Gateway',
    503: 'Service Unavailable', 504: 'Gateway Timeout', 505:
    'HTTP Version Not Supported', 506: 'Variant Also Negotiates', 507:
    'Insufficient Storage', 508: 'Loop Detected', 510: 'Not Extended',
    511: 'Network Authentication Required'
}


def info_banner() -> str:
    """info_banner returns a formatted info banner"""
    return f"{BLUE_BACKGROUND}{BOLD}[INFO]{NORMAL}"


def error_banner() -> str:
    """error_banner returns a formatted error banner"""
    return f"{RED_BACKGROUND}{BOLD}[ERROR]{NORMAL}"


def print_ketter_banner():
    """prints ketter banner to stdout"""
    banner = [10, 32, 95, 32, 32, 32, 32, 32, 32, 32, 32, 95, 32, 32, 32, 95,
              32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 95, 95, 10,
              124, 32, 124, 32, 95, 95, 95, 95, 95, 124, 32, 124, 95, 124, 32,
              124, 95, 32, 95, 95, 95, 32, 95, 32, 95, 95, 32, 32, 92, 32, 92,
              10, 124, 32, 124, 47, 32, 47, 32, 95, 32, 92, 32, 95, 95, 124,
              32, 95, 95, 47, 32, 95, 32, 92, 32, 39, 95, 95, 124, 32, 32, 92,
              32, 92, 10, 124, 32, 32, 32, 60, 32, 32, 95, 95, 47, 32, 124, 95,
              124, 32, 124, 124, 32, 32, 95, 95, 47, 32, 124, 32, 32, 32, 32,
              32, 47, 32, 47, 10, 124, 95, 124, 92, 95, 92, 95, 95, 95, 124,
              92, 95, 95, 124, 92, 95, 95, 92, 95, 95, 95, 124, 95, 124, 32,
              32, 32, 32, 47, 95, 47, 10]

    print(f"{YELLOW}{BOLD}{bytes(banner).decode('utf-8')}{NORMAL}")


def format_user_submitted(user_str: str) -> str:
    return f"{UNDERLINE}{CYAN}{user_str}{NORMAL}"


def capitalise(string: str) -> str:
    pieces = string.split("-")
    capped_pieces = map(lambda piece: piece[0].upper() + piece[1:], pieces)
    return "-".join(capped_pieces)



import aiohttp
import asyncio
import sys
from .utils import *
from .arguments import main as arguments_main
from .downloader import worker


async def main() -> int:
    """executes the program"""

    retval = 0

    headers, urls = arguments_main()

    timeout = aiohttp.ClientTimeout(total=None)
    async with aiohttp.ClientSession(headers=headers, timeout=timeout) as session:
        workers = []

        for idx, url in enumerate(urls):
            workers.append(worker(idx, url, session))

        results = await asyncio.gather(*workers, return_exceptions=True)

        for idx, result in enumerate(results):
            if isinstance(result, Exception):
                retval += 1
                print(
                    f"{error_banner()} {format_user_submitted(urls[idx])}: {result}")

    return retval


def cli():
    print_ketter_banner()
    sys.exit(asyncio.run(main()))


if __name__ == "__main__":
    print_ketter_banner()
    sys.exit(asyncio.run(main()))



"""
arguments is responsible getting, validating and formatting user entered data
to make it ready for use by the program
"""
import argparse
import functools
import os
from .utils import *
from .errors import *
from . import __version__ as VERSION


def create_parser() -> argparse.ArgumentParser:
    """creates and returns a commandline arguments parser"""

    parser = argparse.ArgumentParser(
        description="Asynchronous HTTP downloader", prog="ketter")
    parser.add_argument("-v", "--version", action="version",
                        version=f"{info_banner()} %(prog)s {VERSION}")
    parser.add_argument("--header", action="append",
                        metavar="key=value", help="""custom header to include
                        in all requests. More than one may be specified. May
                        also be used to overwrite automatically generated
                        headers such as the 'User-Agent' header""")
    parser.add_argument("URL_FILE",  help="""text file with urls to be
                        downloaded, separated by newlines. The urls should be
                        written in full, including the url scheme""")

    return parser


def validate_url_file(url_file: str):
    """ensures the url file exists and is a regular file"""

    if not os.path.exists(url_file):
        raise KetterInvalidFileError(
            f"Url file does not exist: {format_user_submitted(url_file)}")

    elif not os.path.isfile(url_file):
        raise KetterInvalidFileError(
            f"Url file not regular file: {format_user_submitted(url_file)}")


def harvest_urls(url_file: str) -> list[str]:
    """reads and formats urls from url file"""

    with open(url_file) as file:
        return [url.strip() for url in file.readlines()]


def harvest_headers(custom_headers: list[str]) -> dict[str, str]:
    """validates and formats passed headers for ready use"""

    def reduce_headers(
            accum_headers: dict[str, str],
            curr_header: str
    ) -> dict[str, str]:
        key, value = curr_header.split("=", maxsplit=1)
        key = key.lower()

        if value == "":
            raise KetterHTTPHeaderError(
                f"Invalid header: {format_user_submitted(curr_header)}")

        accum_headers[key] = value
        return accum_headers

    return functools.reduce(
        reduce_headers,
        custom_headers,
        {}
    )


def main() -> tuple[dict[str, str], list[str]]:
    """returns all commandline arguments ready for use"""

    parser = create_parser()
    args = parser.parse_args()

    try:
        custom_headers = harvest_headers(args.header or [])
        validate_url_file(args.URL_FILE)
        urls = harvest_urls(args.URL_FILE)
    except Exception as e:
        parser.exit(2, f"{error_banner()} {e}")

    headers = {"user-agent": f"ketter/{VERSION}"}
    headers.update(custom_headers)

    def reduce_headers_capitalise(
            accum_headers: dict[str, str],
            curr_header_key: str
    ) -> dict[str, str]:
        accum_headers[capitalise(curr_header_key)] = headers[curr_header_key]
        return accum_headers

    headers = functools.reduce(reduce_headers_capitalise, headers, {})

    return headers, urls



"""errors provides custom errors"""


class KetterHTTPError(Exception):
    """for errors encountered while downloading"""

    ...


class KetterHTTPHeaderError(Exception):
    """for errors relating to custom headers"""

    ...


class KetterInvalidFileError(Exception):
    """for errors relating to url file"""

    ...



"""colour provides ANSI colour codes"""

_ESC = chr(27)

NORMAL = f"{_ESC}[0m"
BOLD = f"{_ESC}[1m"
ITALIC = f"{_ESC}[3m"
UNDERLINE = f"{_ESC}[4m"
BLINK = f"{_ESC}[5m"
INVERT = f"{_ESC}[7m"
INVISIBLE = f"{_ESC}[8m"
STRIKE_THROUGH = f"{_ESC}[9m"
OVERWRITE = f"{_ESC}[2K" + "\r"

BLACK = f"{_ESC}[30m"
RED = f"{_ESC}[31m"
GREEN = f"{_ESC}[32m"
YELLOW = f"{_ESC}[33m"
BLUE = f"{_ESC}[34m"
MAGENTA = f"{_ESC}[35m"
CYAN = f"{_ESC}[36m"
WHITE = f"{_ESC}[37m"

BLACK_BACKGROUND = f"{_ESC}[40m"
RED_BACKGROUND = f"{_ESC}[41m"
GREEN_BACKGROUND = f"{_ESC}[42m"
YELLOW_BACKGROUND = f"{_ESC}[43m"
BLUE_BACKGROUND = f"{_ESC}[44m"
MAGENTA_BACKGROUND = f"{_ESC}[45m"
CYAN_BACKGROUND = f"{_ESC}[46m"
WHITE_BACKGROUND = f"{_ESC}[47m"
